import tempfile
import unittest
import os.path

import itertools
import nltk
from unittest import mock
from gensim import corpora
from requests.exceptions import ConnectionError
import numpy as np

from orangecontrib.text import preprocess
from orangecontrib.text.corpus import Corpus
from orangecontrib.text.preprocess import Preprocessor
from orangecontrib.text.preprocess.normalize import file_to_language, \
    file_to_name, language_to_name, UDPipeModels


def counted(f):
    def wrapped(*args, **kwargs):
        wrapped.calls += 1
        return f(*args, **kwargs)
    wrapped.calls = 0
    return wrapped


class PreprocessTests(unittest.TestCase):
    sentence = "Human machine interface for lab abc computer applications"

    def setUp(self):
        self.corpus = Corpus.from_file('deerwester')

    def test_string_processor(self):
        p = Preprocessor(transformers=preprocess.LowercaseTransformer())
        tokens = p(self.corpus).tokens
        p2 = Preprocessor(transformers=[])
        tokens2 = p2(self.corpus).tokens

        np.testing.assert_equal(tokens,
                                [[t.lower() for t in doc] for doc in tokens2])

        self.assertRaises(TypeError, Preprocessor, string_transformers=1)

    def test_tokenizer(self):
        class SpaceTokenizer(preprocess.BaseTokenizer):
            @classmethod
            def tokenize(cls, string):
                return string.split()
        p = Preprocessor(tokenizer=SpaceTokenizer())

        np.testing.assert_equal(p(self.corpus).tokens,
                         np.array([sent.split() for sent in self.corpus.documents]))

    def test_token_normalizer(self):
        class CapTokenNormalizer(preprocess.BaseNormalizer):
            @classmethod
            def normalize(cls, token):
                return token.capitalize()
        p = Preprocessor(normalizer=CapTokenNormalizer())
        tokens = p(self.corpus).tokens
        p2 = Preprocessor(normalizer=None)
        tokens2 = p2(self.corpus).tokens

        np.testing.assert_equal(
            tokens, [[t.capitalize() for t in doc] for doc in tokens2])

    def test_token_filter(self):
        class SpaceTokenizer(preprocess.BaseTokenizer):
            @classmethod
            def tokenize(cls, string):
                return string.split()

        class LengthFilter(preprocess.BaseTokenFilter):
            @classmethod
            def check(cls, token):
                return len(token) < 4

        p = Preprocessor(tokenizer=SpaceTokenizer(), filters=LengthFilter())
        np.testing.assert_equal(p(self.corpus).tokens,
                         np.array([[token for token in doc.split() if len(token) < 4]
                                   for doc in self.corpus.documents]))

    def test_inplace(self):
        p = Preprocessor(tokenizer=preprocess.RegexpTokenizer('\w'))
        corpus = p(self.corpus, inplace=True)
        self.assertIs(corpus, self.corpus)

        corpus = p(self.corpus, inplace=False)
        self.assertIsNot(corpus, self.corpus)
        self.assertEqual(corpus, self.corpus)

        p = Preprocessor(tokenizer=preprocess.RegexpTokenizer('\w+'))
        corpus = p(self.corpus, inplace=False)
        self.assertIsNot(corpus, self.corpus)
        self.assertNotEqual(corpus, self.corpus)


class TransformationTests(unittest.TestCase):
    def test_call(self):

        class ReverseStringTransformer(preprocess.BaseTransformer):
            name = "Reverse"

            def transform(self, string):
                return string[::-1]

        transformer = ReverseStringTransformer()

        self.assertEqual(transformer('abracadabra'), 'arbadacarba')
        self.assertEqual(transformer(['abra', 'cadabra']), ['arba', 'arbadac'])

        self.assertRaises(TypeError, transformer, 1)

    def test_str(self):
        class ReverseStringTransformer(preprocess.BaseTransformer):
            name = 'reverse'

            def transform(self, string):
                return string[::-1]

        transformer = ReverseStringTransformer()

        self.assertIn('reverse', str(transformer))

    def test_lowercase(self):
        transformer = preprocess.LowercaseTransformer()
        self.assertEqual(transformer.transform('Abra'), 'abra')
        self.assertEqual(transformer.transform('\u00C0bra'), '\u00E0bra')

    def test_strip_accents(self):
        transformer = preprocess.StripAccentsTransformer()
        self.assertEqual(transformer.transform('Abra'), 'Abra')
        self.assertEqual(transformer.transform('\u00C0bra'), 'Abra')

    def test_html(self):
        transformer = preprocess.HtmlTransformer()
        self.assertEqual(transformer('<p>abra<b>cadabra</b><p>'), 'abracadabra')

    def test_url_remover(self):
        url_remover = preprocess.UrlRemover()
        self.assertEqual(url_remover.transform('some link to https://google.com/'), 'some link to ')
        self.assertEqual(url_remover.transform('some link to google.com'), 'some link to google.com')


class TokenNormalizerTests(unittest.TestCase):

    def setUp(self):
        self.stemmer = nltk.PorterStemmer().stem

    def test_str(self):
        stemmer = preprocess.PorterStemmer()
        self.assertIn('porter', str(stemmer).lower())

        stemmer = preprocess.SnowballStemmer('french')
        self.assertIn('french', str(stemmer).lower())

    def test_call(self):
        word = "Testing"
        tokens = ["Testing", "tokenized", "Sentence"]
        stemmer = preprocess.PorterStemmer()
        self.assertEqual(stemmer(word), self.stemmer(word))
        self.assertEqual(stemmer(tokens),
                         [self.stemmer(token) for token in tokens])

    def test_function(self):
        stemmer = preprocess.BaseNormalizer()
        stemmer.normalizer = lambda x: x[:-1]
        self.assertEqual(stemmer.normalize('token'), 'toke')

    def test_snowball(self):
        stemmer = preprocess.SnowballStemmer()
        stemmer.language = 'french'
        token = 'voudrais'
        self.assertEqual(stemmer(token), nltk.SnowballStemmer(language='french').stem(token))

    def test_udpipe(self):
        """Test udpipe token lemmatization"""
        normalizer = preprocess.UDPipeLemmatizer()
        normalizer.language = 'Slovenian'
        self.assertEqual(normalizer('sem'), 'biti')

    def test_udpipe_doc(self):
        """Test udpipe lemmatization with its own tokenization """
        normalizer = preprocess.UDPipeLemmatizer()
        normalizer.language = 'Slovenian'
        normalizer.use_tokenizer = True
        self.assertListEqual(normalizer.normalize_doc('Gori na gori hiša gori'),
                             ['gora', 'na', 'gora', 'hiša', 'goreti'])

    def test_porter_with_bad_input(self):
        stemmer = preprocess.PorterStemmer()
        self.assertRaises(TypeError, stemmer, 10)

    def test_lookup_normalize(self):
        dln = preprocess.DictionaryLookupNormalizer(dictionary={'aka': 'also known as'})
        self.assertEqual(dln.normalize('aka'), 'also known as')


class UDPipeModelsTests(unittest.TestCase):
    def test_label_transform(self):
        """Test helper functions for label transformation"""
        self.assertEqual(file_to_language('slovenian-sst-ud-2.0-170801.udpipe'),
                         'Slovenian sst')
        self.assertEqual(file_to_name('slovenian-sst-ud-2.0-170801.udpipe'),
                         'sloveniansstud2.0170801.udpipe')
        self.assertEqual(language_to_name('Slovenian sst'), 'sloveniansstud')

    def test_udpipe_model(self):
        """Test udpipe models loading from server"""
        models = UDPipeModels()
        self.assertIn('Slovenian', models.supported_languages)
        self.assertEqual(68, len(models.supported_languages))

        local_file = os.path.join(models.local_data,
                                  'slovenian-ud-2.0-170801.udpipe')
        model = models['Slovenian']
        self.assertEqual(model, local_file)
        self.assertTrue(os.path.isfile(local_file))

    def test_udpipe_local_models(self):
        """Test if UDPipe works offline and uses local models"""
        models = UDPipeModels()
        [models.localfiles.remove(f[0]) for f in models.localfiles.listfiles()]
        _ = models['Slovenian']
        with mock.patch('serverfiles.ServerFiles.listfiles',
                        **{'side_effect': ConnectionError()}):
            self.assertIn('Slovenian', UDPipeModels().supported_languages)
            self.assertEqual(1, len(UDPipeModels().supported_languages))

    def test_udpipe_offline(self):
        """Test if UDPipe works offline"""
        self.assertTrue(UDPipeModels().online)
        with mock.patch('serverfiles.ServerFiles.listfiles',
                        **{'side_effect': ConnectionError()}):
            self.assertFalse(UDPipeModels().online)


class FilteringTests(unittest.TestCase):

    def setUp(self):
        self.corpus = Corpus.from_file('deerwester')

    def test_str(self):
        f = preprocess.StopwordsFilter('french')
        self.assertIn('french', str(f).lower())

        f = preprocess.FrequencyFilter(keep_n=None)
        self.assertNotIn('none', str(f).lower())
        f.max_df = .5
        self.assertIn('0.5', str(f))
        f.max_df = .2
        self.assertIn('0.2', str(f))

        f = preprocess.LexiconFilter()
        self.assertIn('lexicon', str(f).lower())

    def test_call(self):
        class DigitsFilter(preprocess.BaseTokenFilter):
            def check(self, token):
                return not token.isdigit()

        df = DigitsFilter()

        self.assertEqual(df([]), [])
        self.assertEqual(df(['a', '1']), ['a'])
        self.assertEqual(df([['a', '1']]), [['a']])

    def test_stopwords(self):
        filter = preprocess.StopwordsFilter('english')

        self.assertFalse(filter.check('a'))
        self.assertTrue(filter.check('filter'))

    def test_lexicon(self):
        filter = preprocess.LexiconFilter(['filter'])
        self.assertFalse(filter.check('false'))
        self.assertTrue(filter.check('filter'))

    def test_keep_n(self):
        ff = preprocess.FrequencyFilter(keep_n=5)
        p = Preprocessor(tokenizer=preprocess.RegexpTokenizer(r'\w+'),
                         filters=[ff])
        processed = p(self.corpus)
        self.assertEqual(len(set(itertools.chain(*processed.tokens))), 5)

    def test_min_df(self):
        ff = preprocess.FrequencyFilter(min_df=.5)
        p = Preprocessor(tokenizer=preprocess.RegexpTokenizer(r'\w+'),
                         filters=[ff])
        processed = p(self.corpus)
        size = len(processed.documents)
        self.assertFrequencyRange(processed, size * .5, size)

        ff.min_df = 2
        processed = p(self.corpus)
        size = len(processed.documents)
        self.assertFrequencyRange(processed, 2, size)

    def test_max_df(self):
        ff = preprocess.FrequencyFilter(max_df=.3)
        p = Preprocessor(tokenizer=preprocess.RegexpTokenizer(r'\w+'),
                         filters=[ff])
        size = len(self.corpus.documents)

        corpus = p(self.corpus)
        self.assertFrequencyRange(corpus, 1, size * .3)

        ff.max_df = 2
        corpus = p(self.corpus)
        self.assertFrequencyRange(corpus, 1, 2)

    def assertFrequencyRange(self, corpus, min_fr, max_fr):
        dictionary = corpora.Dictionary(corpus.tokens)
        self.assertTrue(all(min_fr <= fr <= max_fr
                            for fr in dictionary.dfs.values()))

    def test_word_list(self):
        lexicon = preprocess.LexiconFilter()
        f = tempfile.NamedTemporaryFile()
        f.write(b'hello\nworld\n')
        f.flush()
        lexicon.from_file(f.name)
        self.assertIn('hello', lexicon.lexicon)
        self.assertIn('world', lexicon.lexicon)

    def test_regex_filter(self):
        reg_filter = preprocess.RegexpFilter(r'.')
        filtered = reg_filter(self.corpus.tokens[0])
        self.assertFalse(filtered)

        reg_filter.pattern = 'foo'
        self.assertCountEqual(reg_filter(['foo', 'bar']), ['bar'])

        reg_filter.pattern = '^http'
        self.assertCountEqual(reg_filter(['https', 'http', ' http']), [' http'])

        self.assertFalse(preprocess.RegexpFilter.validate_regexp('?'))
        self.assertTrue(preprocess.RegexpFilter.validate_regexp('\?'))


class TokenizerTests(unittest.TestCase):
    def test_call(self):
        class DashTokenizer(preprocess.BaseTokenizer):
            @classmethod
            def tokenize(cls, string):
                return string.split('-')

        tokenizer = DashTokenizer()
        self.assertEqual(list(tokenizer('dashed-sentence')), ['dashed', 'sentence'])
        self.assertEqual(list(tokenizer(['1-2-3', '-'])), [['1', '2', '3'], ['', '']])

        self.assertRaises(TypeError, tokenizer, 1)

    def test_tokenizer_instance(self):
        class WhitespaceTokenizer(preprocess.BaseTokenizer):
            tokenizer = nltk.WhitespaceTokenizer()
            name = 'whitespace'

        tokenizer = WhitespaceTokenizer()

        sent = "Test \t tokenizer"
        self.assertEqual(tokenizer.tokenize(sent),
                         nltk.WhitespaceTokenizer().tokenize(sent))

    def test_call_with_bad_input(self):
        tokenizer = preprocess.RegexpTokenizer(pattern='\w+')
        self.assertRaises(TypeError, tokenizer.tokenize, 1)
        self.assertRaises(TypeError, tokenizer.tokenize, ['1', 2])

    def test_valid_regexp(self):
        self.assertTrue(preprocess.RegexpTokenizer.validate_regexp('\w+'))

    def test_invalid_regex(self):
        for expr in ['\\', '[', ')?']:
            self.assertFalse(preprocess.RegexpTokenizer.validate_regexp(expr))

    def test_on_change(self):
        tokenizer = preprocess.RegexpTokenizer(pattern=r'\w+')
        tokenizer.on_change = counted(tokenizer.on_change)
        tokenizer.pattern = r'\S+'
        self.assertEqual(tokenizer.on_change.calls, 1)
        self.assertEqual(tokenizer.pattern, r'\S+')

    def test_str(self):
        tokenizer = preprocess.RegexpTokenizer(pattern=r'\S+')
        self.assertIn('\S+', str(tokenizer))

    def test_skip_empty_strings(self):
        tokenizer = preprocess.RegexpTokenizer(pattern=r'[^h ]*')
        tokens = tokenizer('whatever')
        self.assertNotIn('', tokens)



from collections import OrderedDict, Iterable

import tweepy

from Orange import data
from orangecontrib.text import Corpus
from orangecontrib.text.language_codes import code2lang

__all__ = ['Credentials', 'TwitterAPI']


def coordinates_geoJSON(json):
    if json:
        return json.get('coordinates', [None, None])
    return [None, None]


class Credentials:
    """ Twitter API credentials. """

    def __init__(self, consumer_key, consumer_secret):
        self.consumer_key = consumer_key
        self.consumer_secret = consumer_secret
        self.auth = tweepy.OAuthHandler(consumer_key, consumer_secret)
        self._valid = None

    @property
    def valid(self):
        if self._valid is None:
            self.check()
        return self._valid

    def check(self):
        try:
            self.auth.get_authorization_url()
            self._valid = True
        except tweepy.TweepError:
            self._valid = False
        return self._valid

    def __getstate__(self):
        odict = self.__dict__.copy()
        odict['_valid'] = None
        odict.pop('auth')
        return odict

    def __setstate__(self, odict):
        self.__dict__.update(odict)
        self.auth = tweepy.OAuthHandler(self.consumer_key,
                                        self.consumer_secret)

    def __eq__(self, other):
        return isinstance(other, Credentials) \
               and self.consumer_key == other.consumer_key \
               and self.consumer_secret == other.consumer_secret


class TwitterAPI:
    """ Fetch tweets from the Tweeter API.

    Notes:
        Results across multiple searches are aggregated. To remove tweets form
        previous searches and only return results from the last search either
        call `reset` method before searching or provide `collecting=False`
        argument to search method.
    """
    attributes = []
    class_vars = [
        (data.DiscreteVariable('Author'), lambda doc: '@' + doc.author.screen_name),
    ]

    tv = data.TimeVariable('Date')
    metas = [
        (data.StringVariable('Content'), lambda doc: doc.full_text if not
        doc.retweeted else doc.text),
        (tv, lambda doc: TwitterAPI.tv.parse(doc.created_at.isoformat())),
        (data.DiscreteVariable('Language'), lambda doc: doc.lang),
        (data.DiscreteVariable('Location'), lambda doc: getattr(doc.place, 'country_code', None)),
        (data.ContinuousVariable('Number of Likes', number_of_decimals=0),
         lambda doc: doc.favorite_count),
        (data.ContinuousVariable('Number of Retweets', number_of_decimals=0),
         lambda doc: doc.retweet_count),
        (data.DiscreteVariable('In Reply To'),
            lambda doc: '@' + doc.in_reply_to_screen_name if doc.in_reply_to_screen_name else ''),
        (data.DiscreteVariable('Author Name'), lambda doc: doc.author.name),
        (data.StringVariable('Author Description'), lambda doc: doc.author.description),
        (data.ContinuousVariable('Author Statuses Count', number_of_decimals=0),
         lambda doc: doc.author.statuses_count),
        (data.ContinuousVariable('Author Favourites Count', number_of_decimals=0),
         lambda doc: doc.author.favourites_count),
        (data.ContinuousVariable('Author Friends Count', number_of_decimals=0),
         lambda doc: doc.author.friends_count),
        (data.ContinuousVariable('Author Followers Count', number_of_decimals=0),
         lambda doc: doc.author.followers_count),
        (data.ContinuousVariable('Author Listed Count', number_of_decimals=0),
         lambda doc: doc.author.listed_count),
        (data.DiscreteVariable('Author Verified'), lambda doc: str(doc.author.verified)),
        (data.ContinuousVariable('Longitude'),
            lambda doc: coordinates_geoJSON(doc.coordinates)[0]),
        (data.ContinuousVariable('Latitude'),
            lambda doc: coordinates_geoJSON(doc.coordinates)[1]),
    ]

    text_features = [metas[0][0]]       # Content
    string_attributes = [m for m, _ in metas
                         if isinstance(m, data.StringVariable)]

    def __init__(self, credentials,
                 on_progress=None, should_break=None,
                 on_error=None, on_rate_limit=None):
        self.key = credentials
        self.api = tweepy.API(credentials.auth)
        self.container = OrderedDict()
        self.search_history = []

        # Callbacks:
        self.on_error = on_error
        self.on_rate_limit = on_rate_limit
        self.on_progress = on_progress or (lambda *args: args)
        self.should_break = should_break or (lambda *args: False)

    @property
    def tweets(self):
        return self.container.values()

    def search_content(self, content, *, max_tweets=0,
                       lang=None, allow_retweets=True,
                       collecting=False):
        """ Search by content.

        Args:
            content (list of str): A list of key words to search for.
            max_tweets (int): If greater than zero limits the number of
                downloaded tweets.
            lang (str): A language's code (either ISO 639-1 or ISO 639-3
                formats).
            allow_retweets(bool): Whether to download retweets.
            collecting (bool): Whether to collect results across multiple
                search calls.

        Returns:
            Corpus
        """
        if not collecting:
            self.reset()

        if max_tweets == 0:
            max_tweets = float('Inf')

        def build_query():
            nonlocal content
            if not content:
                q = 'from: '
            else:
                if not isinstance(content, list):
                    content = [content]
                q = ' OR '.join(['"{}"'.format(q) for q in content])
            if not allow_retweets:
                q += ' -filter:retweets'
            return q

        query = build_query()
        cursor = tweepy.Cursor(self.api.search, q=query, lang=lang,
                               tweet_mode='extended')
        corpus, count = self.fetch(cursor, max_tweets)
        self.append_history('Content', content, lang if lang else 'Any',
                            str(allow_retweets), count)
        return corpus

    def search_authors(self, authors, *, max_tweets=0, collecting=False):
        """ Search by authors.

        Args:
            authors (list of str): A list of authors to search for.
            max_tweets (int): If greater than zero limits the number of
                downloaded tweets.
            collecting (bool): Whether to collect results across multiple
                search calls.

        Returns:
            Corpus
        """
        if not collecting:
            self.reset()

        if max_tweets == 0:     # set to max allowed for progress
            max_tweets = 3200

        if not isinstance(authors, list):
            authors = [authors]

        cursors = [tweepy.Cursor(self.api.user_timeline, screen_name=a,
                                 tweet_mode='extended')
                   for a in authors]
        corpus, count = self.fetch(cursors, max_tweets)
        self.append_history('Author', authors, None, None, count)
        return corpus

    def fetch(self, cursors, max_tweets):
        if not isinstance(cursors, list):
            cursors = [cursors]

        count = 0
        try:
            for i, cursor in enumerate(cursors):
                for j, tweet in enumerate(cursor.items(max_tweets), start=1):
                    if self.should_break():
                        break
                    if tweet.id not in self.container:
                        count += 1
                    self.container[tweet.id] = tweet
                    if j % 20 == 0:
                        self.on_progress(len(self.container),
                                         (i*max_tweets + j)/
                                         (len(cursors)*max_tweets))
                if self.should_break():
                    break
        except tweepy.TweepError as e:
            if e.response.status_code == 429 and self.on_rate_limit:
                self.on_rate_limit()
            elif self.on_error:
                self.on_error(str(e))
                return None, 0
        return self.create_corpus(), count

    def create_corpus(self):
        return Corpus.from_documents(self.tweets, 'Twitter', self.attributes,
                                     self.class_vars, self.metas,
                                     title_indices=[-1])

    def reset(self):
        """ Removes all downloaded tweets. """
        self.search_history = []
        self.container = OrderedDict()

    def append_history(self, mode, query, lang, allow_retweets, n_tweets):
        query = ', '.join(query) if isinstance(query, Iterable) else query
        if lang in code2lang.keys():
            lang = code2lang[lang]
        self.search_history.append((
            ('Query', query),
            ('Search by', mode),
            ('Language', lang),
            ('Allow retweets', allow_retweets),
            ('Tweets count', n_tweets),
        ))

    def report(self):
        return self.search_history



import nltk
from simhash import Simhash
import numpy as np

from orangecontrib.text.vectorization.base import BaseVectorizer


class SimhashVectorizer(BaseVectorizer):
    name = "Simhash"
    max_f = 1024

    def __init__(self, shingle_len=10, f=64, hashfunc=None):
        """
        Args:
            shingle_len(int): Length of a shingle.
            f(int): Length of a document fingerprints
            hashfunc(callable): A function that accepts a string and returns
                a unsigned integer
        """
        self.f = f
        self._bin_format = '{:0%db}' % self.f
        self.hashfunc = hashfunc
        self.ngram_len = shingle_len

    @staticmethod
    def get_shingles(tokens, n):
        return map(lambda x: ''.join(x), nltk.ngrams(tokens, n))

    def compute_hash(self, tokens):
        return Simhash(self.get_shingles(tokens, self.ngram_len), f=self.f, hashfunc=self.hashfunc).value

    def int2binarray(self, num):
        return [int(x) for x in self._bin_format.format(num)]

    def _transform(self, corpus, source_dict):
        """ Computes simhash values from the given corpus
        and creates a new one with a simhash attribute.

        Args:
            corpus (Corpus): a corpus with tokens.

        Returns:
            Corpus with `simhash` variable
        """

        X = np.array([self.int2binarray(self.compute_hash(doc)) for doc in corpus.tokens], dtype=np.float)
        corpus = corpus.copy()
        corpus.extend_attributes(X,
                                 feature_names=('simhash_{}'.format(int(i) + 1)
                                                for i in range(self.f)),
                                 var_attrs={'hidden': True})
        return corpus

    def report(self):
        return (('Hash length', self.f),
                ('Shingle length', self.ngram_len))



import unittest

from orangecontrib.text.corpus import Corpus
from orangecontrib.text.sentiment import Liu_Hu_Sentiment, Vader_Sentiment


class LiuHuTest(unittest.TestCase):
    def setUp(self):
        self.corpus = Corpus.from_file('deerwester')
        self.method = Liu_Hu_Sentiment('English')
        self.new_cols = 1

    def test_transform(self):
        sentiment = self.method.transform(self.corpus)
        self.assertIsInstance(sentiment, Corpus)
        self.assertEqual(len(sentiment.domain),
                         len(self.corpus.domain) + self.new_cols)

    def test_copy(self):
        sentiment_t = self.method.transform(self.corpus, copy=True)
        self.assertIsNot(self.corpus, sentiment_t)

        sentiment_f = self.method.transform(self.corpus, copy=False)
        self.assertIs(self.corpus, sentiment_f)

    def test_compute_values(self):
        sentiment = self.method.transform(self.corpus)
        computed = Corpus.from_table(sentiment.domain, self.corpus)

        self.assertEqual(sentiment.domain, computed.domain)
        self.assertTrue((sentiment.X == computed.X).all())

    def test_compute_values_to_different_domain(self):
        destination = Corpus.from_file('andersen')

        self.assertFalse(self.corpus.domain.attributes)
        self.assertFalse(destination.domain.attributes)

        sentiment = self.method.transform(self.corpus)
        computed = destination.transform(sentiment.domain)

        self.assertTrue(sentiment.domain.attributes)
        self.assertEqual(sentiment.domain.attributes, computed.domain.attributes)

    def test_empty_corpus(self):
        corpus = Corpus.from_file('deerwester')[:0]
        sentiment = self.method.transform(corpus)
        self.assertEqual(len(sentiment.domain),
                         len(self.corpus.domain) + self.new_cols)
        self.assertEqual(len(sentiment), 0)

class Liu_Hu_Slovenian(unittest.TestCase):
    def setUp(self):
        self.corpus = Corpus.from_file('slo-opinion-corpus')
        self.method = Liu_Hu_Sentiment('Slovenian')
        self.new_cols = 1

    def test_transform(self):
        sentiment = self.method.transform(self.corpus)
        self.assertIsInstance(sentiment, Corpus)
        self.assertEqual(len(sentiment.domain),
                         len(self.corpus.domain) + self.new_cols)

    def test_copy(self):
        sentiment_t = self.method.transform(self.corpus, copy=True)
        self.assertIsNot(self.corpus, sentiment_t)

        sentiment_f = self.method.transform(self.corpus, copy=False)
        self.assertIs(self.corpus, sentiment_f)

    def test_compute_values(self):
        sentiment = self.method.transform(self.corpus)
        computed = Corpus.from_table(sentiment.domain, self.corpus)

        self.assertEqual(sentiment.domain, computed.domain)
        self.assertTrue((sentiment.X == computed.X).all())

    def test_empty_corpus(self):
        corpus = self.corpus[:0]
        sentiment = self.method.transform(corpus)
        self.assertEqual(len(sentiment.domain),
                         len(self.corpus.domain) + self.new_cols)
        self.assertEqual(len(sentiment), 0)


class VaderTest(LiuHuTest):
    def setUp(self):
        self.corpus = Corpus.from_file('deerwester')
        self.method = Vader_Sentiment()
        self.new_cols = 4


if __name__ == "__main__":
    unittest.main()



import os
import re

from Orange.data.io import detect_encoding
from gensim import corpora
from nltk.corpus import stopwords

from orangecontrib.text.misc import wait_nltk_data

__all__ = ['BaseTokenFilter', 'StopwordsFilter', 'LexiconFilter', 'RegexpFilter', 'FrequencyFilter']


class BaseTokenFilter:
    name = NotImplemented

    def __call__(self, corpus):
        if len(corpus) == 0:
            return corpus
        if isinstance(corpus[0], str):
            return self.filter(corpus)
        return [self.filter(tokens) for tokens in corpus]

    def filter(self, tokens):
        return list(filter(self.check, tokens))

    def check(self, token):
        raise NotImplementedError

    def __str__(self):
        return self.name

    def set_up(self):
        """ A method for setting filters up before every __call__. """
        pass

    def tear_down(self):
        """ A method for cleaning up after every __call__. """
        pass


class WordListMixin:
    def __init__(self, word_list=None):
        self.file_path = None
        self.word_list = word_list or []

    def from_file(self, path):
        self.file_path = path
        self.word_list = []
        if not path:
            return

        for encoding in ('utf-8',
                         None,  # sys.getdefaultencoding()
                         detect_encoding(path)):
            try:
                with open(path, encoding=encoding) as f:
                    self.word_list = set(line.strip() for line in f)
            except UnicodeDecodeError:
                continue
            return
        # No encoding worked, raise
        raise UnicodeError("Couldn't determine file encoding")


class StopwordsFilter(BaseTokenFilter, WordListMixin):
    """ Remove tokens present in NLTK's language specific lists or a file. """
    name = 'Stopwords'

    @staticmethod
    @wait_nltk_data
    def supported_languages():
        # get NLTK list of stopwords
        stopwords_listdir = []
        try:
            stopwords_listdir = [file for file in
                                 os.listdir(stopwords._get_root())
                                 if file.islower()]
        except LookupError:  # when no NLTK data is available
            pass

        return [file.capitalize() for file in stopwords_listdir]

    @wait_nltk_data
    def __init__(self, language='English', word_list=None):
        WordListMixin.__init__(self, word_list)
        super().__init__()
        self.language = language

    @property
    def language(self):
        return self._language

    @language.setter
    def language(self, value):
        self._language = value
        if not self._language:
            self.stopwords = []
        else:
            self.stopwords = set(stopwords.words(self.language.lower()))

    def __str__(self):
        config = ''
        config += 'Language: {}, '.format(self.language.capitalize()) if self.language else ''
        config += 'File: {}, '.format(self.file_path) if self.file_path else ''
        return '{} ({})'.format(self.name, config.strip(', '))

    def check(self, token):
        return token not in self.stopwords and token not in self.word_list


class LexiconFilter(BaseTokenFilter, WordListMixin):
    """ Keep only tokens present in a file. """
    name = 'Lexicon'

    def __init__(self, lexicon=None):
        WordListMixin.__init__(self, word_list=lexicon)

    @property
    def lexicon(self):
        return self.word_list

    @lexicon.setter
    def lexicon(self, value):
        self.word_list = set(value)

    def check(self, token):
        return not self.lexicon or token in self.lexicon

    def __str__(self):
        return '{} ({})'.format(self.name, 'File: {}'.format(self.file_path))


class RegexpFilter(BaseTokenFilter):
    """ Remove tokens matching this regular expressions. """
    name = 'Regexp'

    def __init__(self, pattern=r'\.|,|:|!|\?'):
        self._pattern = pattern
        # Compiled Regexes are NOT deepcopy-able and hence to make Corpus deepcopy-able
        # we cannot store then (due to Corpus also storing used_preprocessor for BoW compute values).
        # To bypass the problem regex is compiled before every __call__ and discarded right after.
        self.regex = None
        self.set_up()

    @property
    def pattern(self):
        return self._pattern

    @pattern.setter
    def pattern(self, value):
        self._pattern = value
        self.set_up()

    @staticmethod
    def validate_regexp(regexp):
        try:
            re.compile(regexp)
            return True
        except re.error:
            return False

    def check(self, token):
        return not self.regex.match(token)

    def __str__(self):
        return '{} ({})'.format(self.name, self.pattern)

    def set_up(self):
        """ Compile Regex before the __call__. """
        self.regex = re.compile(self.pattern)

    def tear_down(self):
        """ Delete Regex after every __call__. """
        self.regex = None


class FrequencyFilter(LexiconFilter):
    """Remove tokens with document frequency outside this range;
    use either absolute or relative frequency. """
    name = 'Document frequency'

    def __init__(self, min_df=0., max_df=1., keep_n=None):
        super().__init__()
        self._corpus_len = 0
        self.keep_n = keep_n
        self._max_df = max_df
        self._min_df = min_df

    def fit_filter(self, corpus):
        self._corpus_len = len(corpus)
        tokens = getattr(corpus, 'tokens', corpus)
        dictionary = corpora.Dictionary(tokens)
        dictionary.filter_extremes(self.min_df, self.max_df, self.keep_n)
        self.lexicon = dictionary.token2id.keys()
        return self(tokens), dictionary

    @property
    def max_df(self):
        if isinstance(self._max_df, int):
            return self._max_df / self._corpus_len if self._corpus_len else 1.
        else:
            return self._max_df

    @max_df.setter
    def max_df(self, value):
        self._max_df = value

    @property
    def min_df(self):
        if isinstance(self._min_df, float):
            return int(self._corpus_len * self._min_df) or 1
        else:
            return self._min_df

    @min_df.setter
    def min_df(self, value):
        self._min_df = value

    def __str__(self):
        keep = ', keep {}'.format(self.keep_n) if self.keep_n else ''
        return "{} (range [{}, {}]{})".format(self.name, self._min_df,
                                              self._max_df, keep)

    def check(self, token):
        return token in self.lexicon



"""Country names to ISO3166_alpha2 codes mapping

Roughly generated by the following bash script on GNU/Linux:

    while read cc name; do
        [ ! "$cc" ] &&
            continue
        out=$(isoquery $cc | cut -f3 --complement);
        [ ! "$out" ] &&
            out="$cc"
        [ "$(echo $out | cut -f3)" = "$name" ] &&
            name=''
        echo -e "$out\t$name"    |
            sed -r 's/\s+$//'    |
            sed -r "s/\t/': ['/" |
            sed -r "s/\t/', '/g" |
            sed -r "s/^/'/"      |
            sed -r 's/$/'"'"',],/'
    done < input/cc.list  # cc.list from jVectorMap; format: lines start with ISO3166_alpha2_code else copied as is

Certain details updated by hand.
"""

CC_EUROPE = {
    '_0': ['Kosovo', 'Kosovo, Republic of'],
    '-99': ['N. Cyprus', 'North Cyprus'],
    'AD': ['AND', 'Andorra'],
    'AL': ['ALB', 'Albania'],
    'AT': ['AUT', 'Austria'],
    'AX': ['ALA', 'Åland Islands', 'Aland'],
    'BA': ['BIH', 'Bosnia and Herzegovina', 'Bosnia and Herz.'],
    'BE': ['BEL', 'Belgium'],
    'BG': ['BGR', 'Bulgaria'],
    'BY': ['BLR', 'Belarus'],
    'CH': ['CHE', 'Switzerland'],
    'CY': ['CYP', 'Cyprus'],
    'CZ': ['CZE', 'Czech Republic', 'Czech Rep.'],
    'DE': ['DEU', 'Germany'],
    'DK': ['DNK', 'Denmark'],
    'DZ': ['DZA', 'Algeria'],
    'EE': ['EST', 'Estonia'],
    'EG': ['EGY', 'Egypt'],
    'ES': ['ESP', 'Spain'],
    'FI': ['FIN', 'Finland'],
    'FO': ['FRO', 'Faroe Islands', 'Faeroe Is.'],
    'FR': ['FRA', 'France'],
    'GB': ['GBR', 'United Kingdom', 'Great Britain', 'England'],
    'GE': ['GEO', 'Georgia'],
    'GG': ['GGY', 'Guernsey'],
    'GR': ['GRC', 'Greece'],
    'HR': ['HRV', 'Croatia'],
    'HU': ['HUN', 'Hungary'],
    'IE': ['IRL', 'Ireland'],
    'IL': ['ISR', 'Israel'],
    'IM': ['IMN', 'Isle of Man'],
    'IQ': ['IRQ', 'Iraq'],
    'IS': ['ISL', 'Iceland'],
    'IT': ['ITA', 'Italy'],
    'JE': ['JEY', 'Jersey'],
    'JO': ['JOR', 'Jordan'],
    'LB': ['LBN', 'Lebanon'],
    'LI': ['LIE', 'Liechtenstein'],
    'LT': ['LTU', 'Lithuania'],
    'LU': ['LUX', 'Luxembourg'],
    'LV': ['LVA', 'Latvia'],
    'LY': ['LBY', 'Libya'],
    'MA': ['MAR', 'Morocco'],
    'MD': ['MDA', 'Moldova, Republic of', 'Moldova'],
    'ME': ['MNE', 'Montenegro'],
    'MK': ['MKD', 'Macedonia, Republic of', 'Macedonia'],
    'MT': ['MLT', 'Malta'],
    'NL': ['NLD', 'Netherlands'],
    'NO': ['NOR', 'Norway'],
    'PL': ['POL', 'Poland'],
    'PS': ['PSE', 'Palestine, State of', 'Palestine'],
    'PT': ['PRT', 'Portugal'],
    'RO': ['ROU', 'Romania'],
    'RS': ['SRB', 'Serbia'],
    'RU': ['RUS', 'Russian Federation', 'Russia'],
    'SA': ['SAU', 'Saudi Arabia'],
    'SE': ['SWE', 'Sweden'],
    'SI': ['SVN', 'Slovenia'],
    'SK': ['SVK', 'Slovakia'],
    'SM': ['SMR', 'San Marino'],
    'SY': ['SYR', 'Syrian Arab Republic', 'Syria'],
    'TN': ['TUN', 'Tunisia'],
    'TR': ['TUR', 'Turkey'],
    'UA': ['UKR', 'Ukraine'],
}


CC_WORLD = {
    # Does NOT include CC_EUROPE
    '_1': ['Somaliland'],
    'AE': ['ARE', 'United Arab Emirates'],
    'AF': ['AFG', 'Afghanistan'],
    'AM': ['ARM', 'Armenia'],
    'AO': ['AGO', 'Angola'],
    'AR': ['ARG', 'Argentina'],
    'AU': ['AUS', 'Australia'],
    'AZ': ['AZE', 'Azerbaijan'],
    'BD': ['BGD', 'Bangladesh'],
    'BF': ['BFA', 'Burkina Faso'],
    'BI': ['BDI', 'Burundi'],
    'BJ': ['BEN', 'Benin'],
    'BN': ['BRN', 'Brunei Darussalam', 'Brunei'],
    'BO': ['BOL', 'Bolivia, Plurinational State of', 'Bolivia'],
    'BR': ['BRA', 'Brazil'],
    'BS': ['BHS', 'Bahamas'],
    'BT': ['BTN', 'Bhutan'],
    'BW': ['BWA', 'Botswana'],
    'BZ': ['BLZ', 'Belize'],
    'CA': ['CAN', 'Canada'],
    'CD': ['COD', 'Congo, The Democratic Republic of the', 'Dem. Rep. Congo'],
    'CF': ['CAF', 'Central African Republic', 'Central African Rep.'],
    'CG': ['COG', 'Congo'],
    'CI': ['CIV', "Côte d'Ivoire"],
    'CL': ['CHL', 'Chile'],
    'CM': ['CMR', 'Cameroon'],
    'CN': ['CHN', 'China'],
    'CO': ['COL', 'Colombia'],
    'CR': ['CRI', 'Costa Rica'],
    'CU': ['CUB', 'Cuba'],
    'DJ': ['DJI', 'Djibouti'],
    'DO': ['DOM', 'Dominican Republic', 'Dominican Rep.'],
    'EC': ['ECU', 'Ecuador'],
    'EH': ['ESH', 'Western Sahara', 'W. Sahara'],
    'ER': ['ERI', 'Eritrea'],
    'ET': ['ETH', 'Ethiopia'],
    'FJ': ['FJI', 'Fiji'],
    'FK': ['FLK', 'Falkland Islands [Malvinas]', 'Falkland Is.'],
    'GA': ['GAB', 'Gabon'],
    'GH': ['GHA', 'Ghana'],
    'GL': ['GRL', 'Greenland'],
    'GM': ['GMB', 'Gambia'],
    'GN': ['GIN', 'Guinea'],
    'GQ': ['GNQ', 'Equatorial Guinea', 'Eq. Guinea'],
    'GT': ['GTM', 'Guatemala'],
    'GW': ['GNB', 'Guinea-Bissau'],
    'GY': ['GUY', 'Guyana'],
    'HN': ['HND', 'Honduras'],
    'HT': ['HTI', 'Haiti'],
    'ID': ['IDN', 'Indonesia'],
    'IN': ['IND', 'India'],
    'IR': ['IRN', 'Iran, Islamic Republic of', 'Iran'],
    'JM': ['JAM', 'Jamaica'],
    'JP': ['JPN', 'Japan'],
    'KE': ['KEN', 'Kenya'],
    'KG': ['KGZ', 'Kyrgyzstan'],
    'KH': ['KHM', 'Cambodia'],
    'KP': ['PRK', "Korea, Democratic People's Republic of", 'Dem. Rep. Korea', 'North Korea'],
    'KR': ['KOR', 'Korea, Republic of', 'Korea', 'South Korea'],
    'KW': ['KWT', 'Kuwait'],
    'KZ': ['KAZ', 'Kazakhstan'],
    'LA': ['LAO', "Lao People's Democratic Republic", 'Lao PDR'],
    'LK': ['LKA', 'Sri Lanka'],
    'LR': ['LBR', 'Liberia'],
    'LS': ['LSO', 'Lesotho'],
    'MG': ['MDG', 'Madagascar'],
    'ML': ['MLI', 'Mali'],
    'MM': ['MMR', 'Myanmar'],
    'MN': ['MNG', 'Mongolia'],
    'MR': ['MRT', 'Mauritania'],
    'MW': ['MWI', 'Malawi'],
    'MX': ['MEX', 'Mexico'],
    'MY': ['MYS', 'Malaysia'],
    'MZ': ['MOZ', 'Mozambique'],
    'NA': ['NAM', 'Namibia'],
    'NC': ['NCL', 'New Caledonia'],
    'NE': ['NER', 'Niger'],
    'NG': ['NGA', 'Nigeria'],
    'NI': ['NIC', 'Nicaragua'],
    'NP': ['NPL', 'Nepal'],
    'NZ': ['NZL', 'New Zealand'],
    'OM': ['OMN', 'Oman'],
    'PA': ['PAN', 'Panama'],
    'PE': ['PER', 'Peru'],
    'PG': ['PNG', 'Papua New Guinea'],
    'PH': ['PHL', 'Philippines'],
    'PK': ['PAK', 'Pakistan'],
    'PR': ['PRI', 'Puerto Rico'],
    'PY': ['PRY', 'Paraguay'],
    'QA': ['QAT', 'Qatar'],
    'RW': ['RWA', 'Rwanda'],
    'SB': ['SLB', 'Solomon Islands', 'Solomon Is.'],
    'SD': ['SDN', 'Sudan'],
    'SL': ['SLE', 'Sierra Leone'],
    'SN': ['SEN', 'Senegal'],
    'SO': ['SOM', 'Somalia'],
    'SR': ['SUR', 'Suriname'],
    'SS': ['SSD', 'South Sudan', 'S. Sudan'],
    'SV': ['SLV', 'El Salvador'],
    'SZ': ['SWZ', 'Swaziland'],
    'TD': ['TCD', 'Chad'],
    'TF': ['ATF', 'French Southern Territories', 'Fr. S. Antarctic Lands'],
    'TG': ['TGO', 'Togo'],
    'TH': ['THA', 'Thailand'],
    'TJ': ['TJK', 'Tajikistan'],
    'TL': ['TLS', 'Timor-Leste'],
    'TM': ['TKM', 'Turkmenistan'],
    'TT': ['TTO', 'Trinidad and Tobago'],
    'TW': ['TWN', 'Taiwan, Province of China', 'Taiwan'],
    'TZ': ['TZA', 'Tanzania, United Republic of', 'Tanzania'],
    'UG': ['UGA', 'Uganda'],
    'US': ['USA', 'United States', 'United States of America'],
    'UY': ['URY', 'Uruguay'],
    'UZ': ['UZB', 'Uzbekistan'],
    'VE': ['VEN', 'Venezuela, Bolivarian Republic of', 'Venezuela'],
    'VN': ['VNM', 'Viet Nam', 'Vietnam'],
    'VU': ['VUT', 'Vanuatu'],
    'YE': ['YEM', 'Yemen'],
    'ZA': ['ZAF', 'South Africa'],
    'ZM': ['ZMB', 'Zambia'],
    'ZW': ['ZWE', 'Zimbabwe'],
}


CC_USA = {
    'US-AK': ['AK', 'Alaska'],
    'US-AL': ['AL', 'Alabama'],
    'US-AR': ['AR', 'Arkansas'],
    'US-AZ': ['AZ', 'Arizona'],
    'US-CA': ['CA', 'California'],
    'US-CO': ['CO', 'Colorado'],
    'US-CT': ['CT', 'Connecticut'],
    'US-DC': ['DC', 'District of Columbia'],
    'US-DE': ['DE', 'Delaware'],
    'US-FL': ['FL', 'Florida'],
    'US-GA': ['GA', 'Georgia'],
    'US-HI': ['HI', 'Hawaii'],
    'US-IA': ['IA', 'Iowa'],
    'US-ID': ['ID', 'Idaho'],
    'US-IL': ['IL', 'Illinois'],
    'US-IN': ['IN', 'Indiana'],
    'US-KS': ['KS', 'Kansas'],
    'US-KY': ['KY', 'Kentucky'],
    'US-LA': ['LA', 'Louisiana'],
    'US-MA': ['MA', 'Massachusetts'],
    'US-MD': ['MD', 'Maryland'],
    'US-ME': ['ME', 'Maine'],
    'US-MI': ['MI', 'Michigan'],
    'US-MN': ['MN', 'Minnesota'],
    'US-MO': ['MO', 'Missouri'],
    'US-MS': ['MS', 'Mississippi'],
    'US-MT': ['MT', 'Montana'],
    'US-NC': ['NC', 'North Carolina', 'N. Carolina'],
    'US-ND': ['ND', 'North Dakota', 'N. Dakota'],
    'US-NE': ['NE', 'Nebraska'],
    'US-NH': ['NH', 'New Hampshire'],
    'US-NJ': ['NJ', 'New Jersey'],
    'US-NM': ['NM', 'New Mexico'],
    'US-NV': ['NV', 'Nevada'],
    'US-NY': ['NY', 'New York'],
    'US-OH': ['OH', 'Ohio'],
    'US-OK': ['OK', 'Oklahoma'],
    'US-OR': ['OR', 'Oregon'],
    'US-PA': ['PA', 'Pennsylvania'],
    'US-RI': ['RI', 'Rhode Island'],
    'US-SC': ['SC', 'South Carolina', 'S. Carolina'],
    'US-SD': ['SD', 'South Dakota', 'S. Dakota'],
    'US-TN': ['TN', 'Tennessee'],
    'US-TX': ['TX', 'Texas'],
    'US-UT': ['UT', 'Utah'],
    'US-VA': ['VA', 'Virginia'],
    'US-VT': ['VT', 'Vermont'],
    'US-WA': ['WA', 'Washington'],
    'US-WI': ['WI', 'Wisconsin'],
    'US-WV': ['WV', 'West Virginia'],
    'US-WY': ['WY', 'Wyoming'],
}

US_CITIES = {
    # From https://en.wikipedia.org/wiki/List_of_United_States_cities_by_population
    'US-AK': ['Anchorage'],
    'US-AL': ['Birmingham', 'Montgomery', 'Mobile', 'Huntsville'],
    'US-AR': ['Little Rock'],
    'US-AZ': ['Phoenix', 'Tucson', 'Mesa', 'Chandler', 'Gilbert', 'Glendale', 'Scottsdale', 'Tempe', 'Peoria', 'Surprise'],
    'US-CA': ['Los Angeles', 'San Diego', 'San Jose', 'San Francisco', 'Fresno', 'Sacramento', 'Long Beach', 'Oakland', 'Bakersfield', 'Anaheim', 'Santa Ana', 'Riverside', 'Stockton', 'Chula Vista', 'Irvine', 'Fremont', 'San Bernardino', 'Modesto', 'Oxnard', 'Fontana', 'Moreno Valley', 'Huntington Beach', 'Glendale', 'Santa Clarita', 'Garden Grove', 'Oceanside', 'Rancho Cucamonga', 'Santa Rosa', 'Ontario', 'Elk Grove', 'Corona', 'Lancaster', 'Palmdale', 'Salinas', 'Hayward', 'Pomona', 'Escondido', 'Sunnyvale', 'Torrance', 'Pasadena', 'Orange', 'Fullerton', 'Thousand Oaks', 'Visalia', 'Roseville', 'Concord', 'Simi Valley', 'Santa Clara', 'Victorville', 'Vallejo', 'Berkeley', 'El Monte', 'Downey', 'Costa Mesa', 'Carlsbad', 'Inglewood', 'Fairfield', 'Ventura', 'Temecula', 'Antioch', 'Richmond', 'West Covina', 'Murrieta', 'Norwalk', 'Daly City', 'Burbank', 'Santa Maria', 'El Cajon', 'San Mateo', 'Rialto', 'Clovis', 'East Los Angeles'],
    'US-CO': ['Denver', 'Colorado Springs', 'Aurora', 'Fort Collins', 'Lakewood', 'Thornton', 'Arvada', 'Westminster', 'Pueblo', 'Centennial', 'Boulder'],
    'US-CT': ['Bridgeport', 'New Haven', 'Stamford', 'Hartford', 'Waterbury'],
    'US-DC': ['Washington'],
    'US-FL': ['Jacksonville', 'Miami', 'Tampa', 'Orlando', 'St. Petersburg', 'Hialeah', 'Tallahassee', 'Fort Lauderdale', 'Port St. Lucie', 'Cape Coral', 'Pembroke Pines', 'Hollywood', 'Miramar', 'Gainesville', 'Coral Springs', 'Miami Gardens', 'Clearwater', 'Pompano Beach', 'Palm Bay', 'West Palm Beach', 'Lakeland', 'Brandon'],
    'US-GA': ['Atlanta', 'Columbus', 'Augusta', 'Macon', 'Savannah', 'Athens', 'Sandy Springs'],
    'US-HI': ['Honolulu'],
    'US-IA': ['Des Moines', 'Cedar Rapids', 'Davenport'],
    'US-ID': ['Boise'],
    'US-IL': ['Chicago', 'Aurora', 'Rockford', 'Joliet', 'Naperville', 'Springfield', 'Peoria', 'Elgin'],
    'US-IN': ['Indianapolis', 'Fort Wayne', 'Evansville', 'South Bend'],
    'US-KS': ['Wichita', 'Overland Park', 'Kansas City', 'Olathe', 'Topeka'],
    'US-KY': ['Louisville', 'Lexington'],
    'US-LA': ['New Orleans', 'Baton Rouge', 'Shreveport', 'Lafayette', 'Metairie'],
    'US-MA': ['Boston', 'Worcester', 'Springfield', 'Lowell', 'Cambridge'],
    'US-MD': ['Baltimore'],
    'US-MI': ['Detroit', 'Grand Rapids', 'Warren', 'Sterling Heights', 'Ann Arbor', 'Lansing'],
    'US-MN': ['Minneapolis', 'St. Paul', 'Rochester'],
    'US-MO': ['Kansas City', 'St. Louis', 'Springfield', 'Independence', 'Columbia'],
    'US-MS': ['Jackson'],
    'US-MT': ['Billings'],
    'US-NC': ['Charlotte', 'Raleigh', 'Greensboro', 'Durham', 'Winston–Salem', 'Fayetteville', 'Cary', 'Wilmington', 'High Point'],
    'US-ND': ['Fargo'],
    'US-NE': ['Omaha', 'Lincoln'],
    'US-NH': ['Manchester'],
    'US-NJ': ['Newark', 'Jersey City', 'Paterson', 'Elizabeth'],
    'US-NM': ['Albuquerque', 'Las Cruces'],
    'US-NV': ['Las Vegas', 'Henderson', 'Reno', 'North Las Vegas', 'Paradise', 'Sunrise Manor', 'Spring Valley', 'Enterprise'],
    'US-NY': ['New York City', 'Buffalo', 'Rochester', 'Yonkers', 'Syracuse'],
    'US-OH': ['Columbus', 'Cleveland', 'Cincinnati', 'Toledo', 'Akron', 'Dayton'],
    'US-OK': ['Oklahoma City', 'Tulsa', 'Norman', 'Broken Arrow'],
    'US-OR': ['Portland', 'Salem', 'Eugene', 'Gresham'],
    'US-PA': ['Philadelphia', 'Pittsburgh', 'Allentown'],
    'US-RI': ['Providence'],
    'US-SC': ['Columbia', 'Charleston', 'North Charleston'],
    'US-SD': ['Sioux Falls'],
    'US-TN': ['Memphis', 'Nashville', 'Knoxville', 'Chattanooga', 'Clarksville', 'Murfreesboro'],
    'US-TX': ['Houston', 'San Antonio', 'Dallas', 'Austin', 'Fort Worth', 'El Paso', 'Arlington', 'Corpus Christi', 'Plano', 'Laredo', 'Lubbock', 'Garland', 'Irving', 'Amarillo', 'Grand Prairie', 'Brownsville', 'McKinney', 'Pasadena', 'Frisco', 'Mesquite', 'McAllen', 'Killeen', 'Waco', 'Carrollton', 'Denton', 'Midland', 'Abilene', 'Beaumont', 'Odessa', 'Round Rock', 'Richardson', 'Wichita Falls', 'College Station', 'Pearland', 'Lewisville', 'Tyler'],
    'US-UT': ['Salt Lake City', 'West Valley City', 'Provo', 'West Jordan'],
    'US-VA': ['Virginia Beach', 'Norfolk', 'Chesapeake', 'Richmond', 'Newport News', 'Alexandria', 'Hampton', 'Arlington County'],
    'US-WA': ['Seattle', 'Spokane', 'Tacoma', 'Vancouver', 'Bellevue', 'Kent', 'Everett'],
    'US-WI': ['Milwaukee', 'Madison', 'Green Bay'],
}

EUROPE_CITIES = {
    # As by https://www.countries-ofthe-world.com/capitals-of-europe.html
    '_0': ['Pristina', 'Priština', 'Prishtina', 'Prishtinë'],
    '-99': ['North Nicosia'],
    'AD': ['Andorra la Vella'],
    'AL': ['Tirana'],
    'AT': ['Vienna', 'Wien'],
    'AX': ['Mariehamn'],
    'BA': ['Sarajevo'],
    'BE': ['Brussels', 'Bruxelles', 'Brussel'],
    'BG': ['Sofia'],
    'BY': ['Minsk'],
    'CH': ['Bern', 'Zürich', 'Geneva'],
    'CY': ['Nicosia'],
    'CZ': ['Prague', 'Praha', 'Praga'],
    'DE': ['Berlin', 'Hanover', 'Hamburg', 'Munich', 'München', 'Stuttgart', 'Frankfurt', 'Düsseldorf', 'Cologne', 'Essen', 'Dresden', 'Leipzig'],
    'DK': ['Copenhagen', 'København'],
    'DZ': ['Algiers'],
    'EE': ['Tallinn'],
    'EG': ['Cairo'],
    'ES': ['Madrid', 'Barcelona', 'Murcia', 'Malaga', 'Seville', 'Bilbao', 'Zaragoza', 'Valencia'],
    'FI': ['Helsinki'],
    'FO': ['Tórshavn', 'Torshavn'],
    'FR': ['Paris', 'Nice', 'Montpellier', 'Marseille', 'Lyon', 'Toulouse', 'Bordeaux', 'Nantes'],
    'GB': ['London', 'London Heathrow', 'Manchester', 'Liverpool', 'Bristol', 'Cardiff', 'Oxford', 'Cambridge', 'Glasgow', 'Edinburgh', 'Belfast'],
    'GE': ['Tbilisi'],
    'GG': ['Saint Peter Port'],
    'GR': ['Athens', 'Thessaloniki'],
    'HR': ['Zagreb'],
    'HU': ['Budapest'],
    'IE': ['Dublin'],
    'IL': ['Jerusalem', 'Tel Aviv'],
    'IM': ['Douglas'],
    'IQ': ['Baghdad'],
    'IS': ['Reykjavik'],
    'IT': ['Rome', 'Roma'],
    'JE': ['Saint Helier'],
    'JO': ['Amman'],
    'LB': ['Beirut'],
    'LI': ['Vaduz'],
    'LT': ['Vilnius'],
    'LU': ['Luxembourg'],
    'LV': ['Riga'],
    'LY': ['Tripoli'],
    'MA': ['Rabat', 'Fes', 'Casablanca', 'Marrakesh', 'Marrakech', 'Tangier'],
    'MD': ['Chisinau', 'Chișinău'],
    'ME': ['Podgorica'],
    'MK': ['Skopje'],
    'MT': ['Valletta'],
    'NL': ['Amsterdam', 'The Hague', 'Rotterdam'],
    'NO': ['Oslo', 'Bergen', 'Trondheim'],
    'PL': ['Warsaw', 'Warszawa', 'Krakow', 'Lublin', 'Bialystok', 'Lodz', 'Gdansk', 'Wroclaw'],
    'PT': ['Lisbon', 'Porto'],
    'RO': ['Bucharest', 'București'],
    'RS': ['Belgrade', 'Beograd'],
    'RU': ['Moscow', 'Moskva', 'Saint Petersburg', 'St. Petersburg', 'Kazan', 'Vladivostok', 'Irkutsk', 'Novosibirsk', 'Krasnoyarsk', 'Nizhny Novgorod'],
    'SA': ['Riyadh'],
    'SE': ['Stockholm', 'Gothenburg', 'Malmo'],
    'SI': ['Ljubljana'],
    'SK': ['Bratislava'],
    'SM': ['San Marino'],
    'SY': ['Damascus'],
    'TN': ['Tunis'],
    'TR': ['Ankara', 'Istanbul', 'Constantinople'],
    'UA': ['Kyiv', 'Kiev', 'Kharkiv', 'Odessa', 'Lviv'],
}


WORLD_CITIES = {
    # Capitals and larger cities. Does NOT include CC_EUROPE.
    '_1': ['Hargeisa'],
    'AE': ['Abu Dhabi'],
    'AF': ['Kabul'],
    'AM': ['Yerevan'],
    'AO': ['Luanda'],
    'AR': ['Buenos Aires'],
    'AU': ['Canberra', 'Sydney', 'Melbourne', 'Cairns', 'Adelaide', 'Perth'],
    'AZ': ['Baku'],
    'BD': ['Dhaka'],
    'BF': ['Ouagadougou'],
    'BI': ['Bujumbura'],
    'BJ': ['Porto-Novo', 'Porto Novo', 'Cotonou'],
    'BN': ['Bandar Seri Begawani'],
    'BO': ['La Paz', 'Sucre'],
    'BR': ['Brasília', 'Brasilia', 'Sao Paulo', 'Sao Paulo', 'Rio de Janeiro'],
    'BS': ['Nassau'],
    'BT': ['Thimphu'],
    'BW': ['Gaborone'],
    'BZ': ['Belmopan'],
    'CA': ['Ottawa', 'Vancouver', 'Calgary', 'Edmonton', 'Winnipeg', 'Toronto', 'Montreal', 'Québec', 'Quebec'],
    'CD': ['Kinshasa'],
    'CF': ['Bangui'],
    'CG': ['Brazzaville'],
    'CI': ['Yamoussoukro', 'Abidjan'],
    'CL': ['Santiago'],
    'CM': ['Yaoundé', 'Yaounde'],
    'CN': ['Beijing'],
    'CO': ['Bogotá', 'Bogota'],
    'CR': ['San José', 'San Jose'],
    'CU': ['Havana'],
    'DJ': ['Djibouti'],
    'DO': ['Santo Domingo'],
    'EC': ['Quito'],
    'EH': ['Laayoune'],
    'ER': ['Asmara'],
    'ET': ['Addis Ababa'],
    'FJ': ['Suva'],
    'FK': ['Stanley'],
    'GA': ['Libreville'],
    'GH': ['Accra'],
    'GL': ['Nuuk'],
    'GM': ['Banjul'],
    'GN': ['Conakry'],
    'GQ': ['Malabo'],
    'GT': ['Guatemala City'],
    'GW': ['Bissau'],
    'GY': ['Georgetown'],
    'HN': ['Tegucigalpa'],
    'HT': ['Port-au-Prince', 'Port au Prince'],
    'ID': ['Jakarta'],
    'IN': ['New Delhi', 'Mumbai', 'Bombay', 'Bengaluru', 'Hyderabad'],
    'IR': ['Tehran'],
    'JM': ['Kingston'],
    'JP': ['Tokyo', 'Osaka', 'Hiroshima', 'Fukuoka', 'Sapporo'],
    'KE': ['Nairobi'],
    'KG': ['Bishkek'],
    'KH': ['Phnom Penh'],
    'KP': ['Pyongyang'],
    'KR': ['Seoul'],
    'KW': ['Kuwait City'],
    'KZ': ['Astana'],
    'LA': ['Vientiane'],
    'LK': ['Sri Jayawardenepura Kotte', 'Colombo'],
    'LR': ['Monrovia'],
    'LS': ['Maseru'],
    'MG': ['Antananarivo'],
    'ML': ['Bamako'],
    'MM': ['Naypyidaw'],
    'MN': ['Ulaanbaatar', 'Ulan Bator'],
    'MR': ['Nouakchott'],
    'MW': ['Lilongwe'],
    'MX': ['Mexico City'],
    'MY': ['Kuala Lumpur', 'Putrajaya'],
    'MZ': ['Maputo'],
    'NA': ['Windhoek'],
    'NC': ['Nouméa', 'Noumea'],
    'NE': ['Niamey'],
    'NG': ['Abuja'],
    'NI': ['Managua'],
    'NP': ['Kathmandu'],
    'NZ': ['Wellington', 'Auckland', 'Christchurch', 'Queenstown'],
    'OM': ['Muscat'],
    'PA': ['Panama City'],
    'PE': ['Lima'],
    'PG': ['Port Moresby'],
    'PH': ['Manila'],
    'PK': ['Islamabad'],
    'PR': ['San Juan'],
    'PY': ['Asunción', 'Asuncion'],
    'QA': ['Doha'],
    'RW': ['Kigali'],
    'SB': ['Honiara'],
    'SD': ['Khartoum'],
    'SL': ['Freetown'],
    'SN': ['Dakar'],
    'SO': ['Mogadishu'],
    'SR': ['Paramaribo'],
    'SS': ['Juba'],
    'SV': ['San Salvador'],
    'SZ': ['Mbabane', 'Lobamba'],
    'TD': ['N\'Djamena', 'NDjamena', 'N Djamena'],
    'TG': ['Lomé', 'Lome'],
    'TH': ['Bangkok'],
    'TJ': ['Dushanbe'],
    'TL': ['Dili'],
    'TM': ['Ashgabat'],
    'TT': ['Port of Spain'],
    'TW': ['Taipei'],
    'TZ': ['Dodoma'],
    'UG': ['Kampala'],
    'US': ['Washington, D.C.', 'Washington D.C.'],
    'UY': ['Montevideo'],
    'UZ': ['Tashkent'],
    'VE': ['Caracas'],
    'VN': ['Hanoi', 'Ho Chi Minh'],
    'VU': ['Port Vila'],
    'YE': ['Sana\'a', 'Sanaa'],
    'ZA': ['Pretoria', 'Bloemfontein', 'Cape Town', 'Johannesburg'],
    'ZM': ['Lusaka'],
    'ZW': ['Harare'],
}


# Extend USA with US_CITIES
for i in US_CITIES: CC_USA[i].extend(US_CITIES[i])

# Extend world with Europe and USA
CC_WORLD.update(CC_EUROPE)
CC_WORLD['US'].extend(i for v in CC_USA.values() for i in v[1:])

# Extend WORLD with WORLD_CITIES
for i in WORLD_CITIES: CC_WORLD[i].extend(WORLD_CITIES[i])

# Extend CC_EUROPE with EUROPE_CITIES
for i in EUROPE_CITIES: CC_EUROPE[i].extend(EUROPE_CITIES[i])


def _invert_mapping(dict):
    return {(v.lower() if len(v) > 3 else v): k for k in dict for v in dict[k]}


INV_CC_EUROPE = _invert_mapping(CC_EUROPE)
INV_CC_WORLD = _invert_mapping(CC_WORLD)
INV_CC_USA = _invert_mapping(CC_USA)

SET_CC_EUROPE = set(INV_CC_EUROPE.keys()) | set(INV_CC_EUROPE.values())
SET_CC_USA = set(INV_CC_USA.keys()) | set(INV_CC_USA.values())



from gensim import models

from .topics import GensimWrapper

models.LsiModel.update = models.LsiModel.add_documents
models.LsiModel.add_documents = lambda self, *args, **kwargs: self.update(*args, **kwargs)


class LsiWrapper(GensimWrapper):
    name = 'Latent Semantic Indexing'
    Model = models.LsiModel
    has_negative_weights = True
